<?php
/***************************************************************
*  Copyright notice
*
*  (c) 2013 IMIA net based solutions (d.frerich@imia.de)
*  All rights reserved
*
*  This script is part of the TYPO3 project. The TYPO3 project is
*  free software; you can redistribute it and/or modify
*  it under the terms of the GNU General Public License as published by
*  the Free Software Foundation; either version 2 of the License, or
*  (at your option) any later version.
*
*  The GNU General Public License can be found at
*  http://www.gnu.org/copyleft/gpl.html.
*
*  This script is distributed in the hope that it will be useful,
*  but WITHOUT ANY WARRANTY; without even the implied warranty of
*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*  GNU General Public License for more details.
*
*  This copyright notice MUST APPEAR in all copies of the script!
***************************************************************/

namespace IMIA\BaseExt\Utility;

use Symfony\Component\Yaml\Yaml;
use Doctrine\Common\Annotations\AnnotationReader;
use IMIA\BaseExt\Metadata\Driver\AnnotationDriver;
use IMIA\BaseExt\Utility\DynamicConfiguration\SQL;
use IMIA\BaseExt\Utility\DynamicConfiguration\TCA;
use IMIA\BaseExt\Utility\DynamicConfiguration\TypoScript;

/**
 * @package     imia_base_ext
 * @subpackage  Utility
 * @author      David Frerich <d.frerich@imia.de>
 */
class DynamicConfiguration
{
    /**
     * @var Cache
     */
    static public $cache;

    /**
     * @var string
     */
    static public $cacheKey;

    /**
     * @var boolean
     */
    static private $create = false;

    /**
     * @var array
     */
    static private $settings = array();

    /**
     * @var array
     */
    static private $metadata = array();

    /**
     * @var array
     */
    static private $autoload = array();

    /**
     * @var array
     */
    static private $ignoreAnnotations = array('validate', 'lazy', 'inject', 'fixme', 'todo');

    /**
     * Build the config and autoload for each extension
     */
    static public function build($type = 'localconf')
    {
        self::load();

        foreach (self::$settings as $extKey => $settings) {
            if (self::$create) {
                if ($settings['autoload'] != false) {
                    self::createAutoloading($extKey);
                }
            }

            self::includeCached($extKey, $type);
        }

        if (count(self::$autoload)) {
            self::updateAutoloadingCache(self::$autoload);
        }

        if (self::$create) {
            @exec('chmod 777 -R ' . PATH_site . 'typo3temp');
        }
    }
    /**
     * Build the sql, typoscript and tca for an extension
     */
    static public function create($extKey)
    {
        self::load();
        self::clearCache($extKey);
        self::createSQL($extKey);
        self::createTCA($extKey);
        self::createTypoScript($extKey);
    }

    /**
     * @return array
     */
    static public function getCacheActions()
    {
        self::load();

        $cacheActions = array();
        foreach (self::$settings as $extKey => $settings) {
            $cmd = 'imia_base_ext-' . $extKey;
			$title = sprintf($GLOBALS['LANG']->sL('LLL:EXT:imia_base_ext/Resources/Private/Language/locallang.xlf:rm.clearCacheMenu_imiabaseext', TRUE), $extKey);

			$cacheActions[] = array(
				'id'    => $cmd,
				'title' => $title,
				'href'  => 'tce_db.php?vC=' . $GLOBALS['BE_USER']->veriCode() .
                    '&cacheCmd=' . $cmd . '&ajaxCall=1' . \TYPO3\CMS\Backend\Utility\BackendUtility::getUrlToken('tceAction'),
				'icon'  => \TYPO3\CMS\Backend\Utility\IconUtility::getSpriteIcon('actions-system-cache-clear-impact-low')
			);
        }

        return $cacheActions;
    }

    /**
     * Clear the cached settings file
     */
    static public function clearCache($extKey = null)
    {
        self::init();

        if ($extKey) {
            $extPath = \TYPO3\CMS\Core\Utility\ExtensionManagementUtility::extPath($extKey, 'Configuration');

            if (!is_dir($extPath)) {
                @mkdir($extPath);
            }
            self::rrmdir($extPath . '/Cache');
            mkdir($extPath . '/Cache');
            mkdir($extPath . '/Cache/SQL');
            mkdir($extPath . '/Cache/TCA');
            mkdir($extPath . '/Cache/TypoScript');

            @exec('chmod 777 -R ' . $extPath . '/Cache');
        } else {
            self::$cache->resetCache('cache_core/ext_settings_' . self::$cacheKey . '.php');
        }
    }

    static private function rrmdir($dir)
    {
        foreach (glob($dir . '/*') as $file) {
            if(is_dir($file)) {
                self::rrmdir($file);
            } else {
                unlink($file);
            }
        }

        rmdir($dir);
    }

    /**
     * @param string $extKey
     * @param string $type
     */
    static private function includeCached($extKey, $type)
    {
        switch ($type) {
            case 'localconf':
                $file = self::getTyposcript($extKey);
                if ($file->exists()) {
                    @require_once($file->getFilePath());
                }
                break;
            case 'tables':
                $file = self::getTCA($extKey);
                if ($file->exists()) {
                    @require_once($file->getFilePath());
                }
                break;
        }
    }

    /**
     * @param string $extKey
     * @return File
     */
    static public function getSQL($extKey)
    {
        return new File('ext_tables.sql', \TYPO3\CMS\Core\Utility\ExtensionManagementUtility::extPath($extKey, 'Configuration/Cache/SQL/'));
    }

    /**
     * @param string $extKey
     * @return File
     */
    static private function getTCA($extKey)
    {
        return new File('ext_tables.php', \TYPO3\CMS\Core\Utility\ExtensionManagementUtility::extPath($extKey, 'Configuration/Cache/TCA/'));
    }

    /**
     * @param string $extKey
     * @return File
     */
    static private function getTyposcript($extKey)
    {
        return new File('ext_localconf.php', \TYPO3\CMS\Core\Utility\ExtensionManagementUtility::extPath($extKey, 'Configuration/Cache/TypoScript/'));
    }

    /**
     * Load the extension settings
     */
    static private function init()
    {
        self::$cacheKey = sha1(TYPO3_version . PATH_site . 'imia_base_ext');

        if (!self::$cache) {
            self::$cache = new Cache(PATH_site . 'typo3temp/Cache/Code/');
        }
    }

    /**
     * Load the extension settings
     */
    static private function load()
    {
        self::init();

        if (self::$cache->hasCache('cache_core/ext_settings_' . self::$cacheKey . '.php')) {
            self::$settings = self::$cache->loadCache('cache_core/ext_settings_' . self::$cacheKey . '.php');
        } else {
            foreach ($GLOBALS['TYPO3_LOADED_EXT'] as $extKey => $extInfo) {
                $extPath = \TYPO3\CMS\Core\Utility\ExtensionManagementUtility::extPath($extKey);
                $settingsPath = $extPath . 'ext_baseext.yml';

                // deprecated
                if (!file_exists($settingsPath)) {
                    $settingsPath = $extPath . 'Configuration/ImiaBaseExt/settings.yml';
                }

                if (file_exists($settingsPath)) {
                    self::$settings[$extKey] = Yaml::parse($settingsPath);
                    self::$settings[$extKey]['path'] = $extPath;
                }
            }

            self::$cache->writeCache('cache_core/ext_settings_' . self::$cacheKey . '.php', self::$settings);
            self::$create = true;
        }
    }

    /**
     * @param string $extKey
     */
    static private function createSQL($extKey)
    {
        $cacheFile = self::getSQL($extKey);
        if (!$cacheFile->exists()) {
            $metadata = self::loadMetadata($extKey);

            $sqlGen = new SQL($extKey);
            $sql = $sqlGen->build($metadata);

            $cacheFile->write($sql);
        }
    }

    /**
     * @param string $extKey
     */
    static private function createTCA($extKey)
    {
        $cacheFile = self::getTCA($extKey);
        if (!$cacheFile->exists()) {
            $cacheFile->setContents('<?php' . LF .
                'if (!defined (\'TYPO3_MODE\')) {' . LF .
                '	die (\'Access denied.\');' . LF .
                '}' . LF);

            $metadata = self::loadMetadata($extKey);

            $tcaGen = new TCA($extKey);
            $cacheFiles = $tcaGen->build($metadata, $cacheFile);

            foreach ($cacheFiles as $file) {
                $file->write();
            }
        }
    }

    /**
     * @param string $extKey
     */
    static private function createTypoScript($extKey)
    {
        $cacheFile = self::getTyposcript($extKey);
        $tsCacheFile = new File('setup.ts', \TYPO3\CMS\Core\Utility\ExtensionManagementUtility::extPath($extKey, 'Configuration/Cache/TypoScript/'));

        if (!$tsCacheFile->exists() || !$cacheFile->exists()) {
            $cacheFile->setContents('<?php' . LF .
                'if (!defined (\'TYPO3_MODE\')) {' . LF .
                '	die (\'Access denied.\');' . LF .
                '}' . LF);

            $metadata = self::loadMetadata($extKey);

            $tsGen = new TypoScript($extKey);
            $typoScript = $tsGen->build($metadata);

            $tsCacheFile->setContents($typoScript);
            $tsCacheFile->write();

            $cacheFile->addContents('\TYPO3\CMS\Core\Utility\ExtensionManagementUtility::addTypoScript($_EXTKEY, \'setup\', file_get_contents(\'' . $tsCacheFile->getFilePath() . '\'), true);');
            $cacheFile->write();
        }
    }

    /**
     * @param string $extKey
     */
    static private function createAutoloading($extKey)
    {
        $extPath = \TYPO3\CMS\Core\Utility\ExtensionManagementUtility::extPath($extKey);
        self::$autoload = array_merge(self::$autoload, self::getClassesFromPath($extPath . 'Classes'));
    }

    static private function getClassesFromPath($path)
    {
        $classes = array();
        if (is_dir($path)) {
            $dir = dir($path);
            while (false !== ($entry = $dir->read())) {
                if (substr($entry, strlen($entry)-4) == '.php') {
                    $classes[strtolower(self::getClassFromPath($path . '/' . $entry))] = $path . '/' . $entry;
                } elseif ($entry != '.' && $entry != '..') {
                    $classes = array_merge($classes, self::getClassesFromPath($path . '/' . $entry));
                }
            }
            $dir->close();
        }

        return $classes;
    }

    /**
     * Update the typo3 loaded extension information array cache
     */
    static private function updateTypo3LoadedExtensionInformation()
    {
        \TYPO3\CMS\Core\Utility\ExtensionManagementUtility::removeCacheFiles();

        $cacheIdentifier = 'loaded_extensions_' . sha1((TYPO3_version . PATH_site . 'loadedExtensions'));
        $codeCache = $GLOBALS['typo3CacheManager']->getCache('cache_core');
        $codeCache->set($cacheIdentifier, 'return ' . var_export($GLOBALS['TYPO3_LOADED_EXT'], TRUE) . ';');
    }


	/**
	 * Set or update class loader cache entry.
     *
     * @param array $additionalClassRegistry
	 */
	static private function updateAutoloadingCache($additionalClassRegistry)
    {
        \TYPO3\CMS\Core\Core\ClassLoader::unregisterAutoloader();

        $identifier = 'ClassLoader_' . sha1((TYPO3_version . PATH_site . 'ClassLoader'));

        $phpCodeCache = $GLOBALS['typo3CacheManager']->getCache('cache_core');
		if ($phpCodeCache->has($identifier)) {
			list($classRegistry, $aliasToClassNameMapping) = require(PATH_site . 'typo3temp/Cache/Code/cache_core/' . $identifier . '.php');
		}
        $classRegistry = array_merge($classRegistry, $additionalClassRegistry);

		$cachedFileContent = 'return ' . var_export(array($classRegistry, $aliasToClassNameMapping), true) . ';';
		$GLOBALS['typo3CacheManager']->getCache('cache_core')->set($identifier, $cachedFileContent);

        \TYPO3\CMS\Core\Core\ClassLoader::registerAutoloader();
	}

    /**
     * @param string $extKey
     * @return array
     */
    static private function loadMetadata($extKey)
    {
        if (!array_key_exists($extKey, self::$metadata)) {
            $classes = self::loadClasses(self::$settings[$extKey]['path'] . '/Classes/Domain/Model');
            $classNames = array();
            foreach ($classes as $classPath) {
                @require_once($classPath);
                $classNames[] = self::getClassFromPath($classPath);
            }

            $reader = new AnnotationReader();
            foreach (self::$ignoreAnnotations as $annotation) {
                $reader::addGlobalIgnoredName(strtolower($annotation));
                $reader::addGlobalIgnoredName(strtoupper($annotation));
            }

            $annotationDriver = new AnnotationDriver($reader);
            $metadataFactory = new \Metadata\MetadataFactory($annotationDriver);

            self::$metadata[$extKey] = array();
            foreach ($classNames as $className) {
                self::$metadata[$extKey][$className] = $metadataFactory->getMetadataForClass($className);
            }
        }

        return self::$metadata[$extKey];
    }

    /**
     * @param string $path
     * @return array
     */
    static private function loadClasses($path)
    {
        $classes = array();

        if (is_dir($path)) {
            $dir = dir($path);
            while (false !== ($entry = $dir->read())) {
                if (!in_array($entry, array('.', '..'))) {
                    $entryPath = realpath($path . '/' . $entry);
                    if (is_dir($entryPath)) {
                        $classes = array_merge($classes, self::loadClasses($entryPath));
                    } elseif (preg_match('/\.php$/ism', $entry) && is_file($entryPath)) {
                        $classes[] = $entryPath;
                    }
                }
            }
            $dir->close();
        }

        return $classes;
    }

    /**
     * @param string $path
     * @return string
     */
    static private function getClassFromPath($path)
    {
        $className = self::getClassFromPathByToken($path);

        if (!$className) {
            die('No class found in: ' . $path);
        } elseif (substr($className, strlen($className)-1) == '\\') {
            $className .= preg_replace('/^.*?([^\/]+)\.php$/ism', '$1', $path);
        }

        return $className;
    }

    /**
     * @param string $path
     * @return string
     */
    static private function getClassFromPathByToken($path)
    {
        $class = '';
        $namespace = '';
        $buffer = '';

        $fp = fopen($path, 'r');
        $i = 0;
        while (!feof($fp) && !$class) {
            $buffer .= fread($fp, 512);
            $tokens = @token_get_all($buffer);

            if (strpos($buffer, '{') === false) continue;

            for (;$i<count($tokens);$i++) {
                if ($tokens[$i][0] === T_NAMESPACE) {
                    for ($j=$i+1;$j<count($tokens); $j++) {
                        if ($tokens[$j][0] === T_STRING) {
                             $namespace .= '\\'.$tokens[$j][1];
                        } else if ($tokens[$j] === '{' || $tokens[$j] === ';') {
                             break;
                        }
                    }
                }

                if ($tokens[$i][0] === T_CLASS) {
                    for ($j=$i+1;$j<count($tokens);$j++) {
                        if ($tokens[$j] === '{') {
                            $class = $tokens[$i+2][1];
                        }
                    }
                }
            }
        }

        return substr($namespace, 1) . '\\' . $class;
    }
}