<?php
/***************************************************************
*  Copyright notice
*
*  (c) 2013 IMIA net based solutions (d.frerich@imia.de)
*  All rights reserved
*
*  This script is part of the TYPO3 project. The TYPO3 project is
*  free software; you can redistribute it and/or modify
*  it under the terms of the GNU General Public License as published by
*  the Free Software Foundation; either version 2 of the License, or
*  (at your option) any later version.
*
*  The GNU General Public License can be found at
*  http://www.gnu.org/copyleft/gpl.html.
*
*  This script is distributed in the hope that it will be useful,
*  but WITHOUT ANY WARRANTY; without even the implied warranty of
*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*  GNU General Public License for more details.
*
*  This copyright notice MUST APPEAR in all copies of the script!
***************************************************************/

namespace IMIA\BaseExt\Xclass\Backend\Form;

/**
 * @package     imia_base_ext
 * @subpackage  ViewHelpers
 * @author      David Frerich <d.frerich@imia.de>
 */
class ElementConditionMatcher extends \TYPO3\CMS\Backend\Form\ElementConditionMatcher
{
	/**
	 * @var string
	 */
	protected $recordTable = '';

	/**
	 * @param string $recordTable
	 * @return ElementConditionMatcher
	 */
	public function setRecordTable($recordTable)
	{
		$this->recordTable = $recordTable;

		return $this;
	}

	/**
	 * @param string $fieldName
	 * @param array $record
	 * @return string
	 */
	protected function getRecordFieldValue($fieldName, $record, $recordTable = '')
	{
		if (strpos($fieldName, '.') !== false) {
			$fieldValue = null;

			$fieldNameParts = explode('.', $fieldName, 2);
			$parentFieldName = $fieldNameParts[0];
			$subFieldName = $fieldNameParts[1];

			if (array_key_exists($parentFieldName, $record) && $record[$parentFieldName]) {
				$recordFieldValue = $record[$parentFieldName];

				if ($recordTable) {
					$fieldTCA = $GLOBALS['TCA'][$recordTable]['columns'][$parentFieldName];

					$subRecordTable = '';
					$subRecordUid = 0;
					switch ($fieldTCA['config']['type']) {
						case 'select':
							break;
						case 'group':
							if ($fieldTCA['config']['internal_type'] == 'db') {
								$recordFieldValues = explode(',', $recordFieldValue);
								$allowedTables = array_map('trim', explode(',', $fieldTCA['config']['allowed']));

								if (count($allowedTables) === 1 || is_numeric($recordFieldValues)) {
									$subRecordTable = array_shift($allowedTables);
								} else {
									foreach ($allowedTables as $table) {
										if (strpos($recordFieldValues[0], $table) === 0) {
											$subRecordTable = $table;
											break;
										}
									}
								}

								$subRecordUid = (int)array_shift(explode('|', str_replace($subRecordTable . '_', '', $recordFieldValue)));
							}
							break;
						case 'inline':
							break;
					}

					if ($subRecordTable && $subRecordUid) {
						$res = $GLOBALS['TYPO3_DB']->exec_SELECTquery('*', $subRecordTable, 'uid=' . $subRecordUid . \TYPO3\CMS\Backend\Utility\BackendUtility::deleteClause($subRecordTable));
						$subRecord = $GLOBALS['TYPO3_DB']->sql_fetch_assoc($res);
						$GLOBALS['TYPO3_DB']->sql_free_result($res);

						if ($subRecord) {
							\TYPO3\CMS\Backend\Utility\BackendUtility::fixVersioningPid($subRecordTable, $subRecord);
							$fieldValue = $this->getRecordFieldValue($subFieldName, $subRecord, $subRecordTable);
						}
					}
				}
			}
		} else {
			$fieldValue = $record[$fieldName];
		}

		return $fieldValue;
	}

	/**
	 * @param string $condition
	 * @return boolean
	 */
	protected function matchFieldCondition($condition)
    {
		list($fieldName, $operator, $operand) = explode(':', $condition, 3);
		if ($this->flexformValueKey) {
			if (strpos($fieldName, 'parentRec.') !== FALSE) {
				$fieldNameParts = explode('.', $fieldName, 2);
				$fieldValue = $this->record['parentRec'][$fieldNameParts[1]];
			} else {
				$fieldValue = $this->record[$fieldName][$this->flexformValueKey];
			}
		} else {
			$fieldValue = $this->getRecordFieldValue($fieldName, $this->record, $this->recordTable);
		}

		$result = FALSE;
		switch ($operator) {
			case 'REQ':
				if (strtoupper($operand) === 'TRUE') {
					$result = (bool) $fieldValue;
				} else {
					$result = !$fieldValue;
				}
				break;
			case '>':
				$result = $fieldValue > $operand;
				break;
			case '<':
				$result = $fieldValue < $operand;
				break;
			case '>=':
				$result = $fieldValue >= $operand;
				break;
			case '<=':
				$result = $fieldValue <= $operand;
				break;
			case '-':
			case '!-':
				list($minimum, $maximum) = explode('-', $operand);
				$result = $fieldValue >= $minimum && $fieldValue <= $maximum;
				if ($operator{0} === '!') {
					$result = !$result;
				}
				break;
			case 'IN':
			case '!IN':
			case '=':
			case '!=':
				$result = \TYPO3\CMS\Core\Utility\GeneralUtility::inList($operand, $fieldValue);
				if ($operator{0} === '!') {
					$result = !$result;
				}
				break;
		}

		return $result;
	}
}